import configparser
import os
import sys
import textwrap
import tomllib
from typing import Any, Optional

import jsonschema

from rust2rpm import log


TOML_SCHEMA = {
    "type": "object",
    "properties": {
        # package properties
        "package": {
            "type": "object",
            "properties": {
                # override package Summary
                "summary": {
                    "type": "string",
                },
                # override package description
                "description": {
                    "type": "string",
                },
                # supported target architectures
                "supported-arches": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
            },
            "additionalProperties": False,
        },
        # configure which tests are run
        "tests": {
            "type": "object",
            "properties": {
                "run": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "enum": ["none", "all", "lib", "bin", "doc", "bins", "tests"],
                        "uniqueItems": True,
                    },
                },
                "skip": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                "skip-exact": {
                    "type": "boolean",
                },
                "comments": {
                    "type": "array",
                    "items": {
                        "type": "string",
                    },
                },
            },
            "additionalProperties": False,
            "required": ["comments"],
        },
        # feature flags passed to cargo
        "features": {
            "type": "object",
            "properties": {
                # pass "--all-features"
                "enable-all": {
                    "type": "boolean",
                },
                # pass "--features foo,bar"
                "enable": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # hide feature subpackages
                "hide": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
            },
            "additionalProperties": False,
        },
        # additional RPM Build/Requires
        "requires": {
            "type": "object",
            "properties": {
                # additional BuildRequires
                "build": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # additional BuildRequires gated by the "check" bconf
                "test": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # additional Requires for the -devel subpackage
                "lib": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # additional Requires for the binary subpackage
                "bin": {
                    "type": "array",
                    "items": {
                        "type": "string",
                        "uniqueItems": True,
                    },
                },
                # additional Requires for feature subpackages
                "features": {
                    "type": "object",
                    "patternProperties": {
                        "^.*$": {
                            "type": "array",
                            "items": {
                                "type": "string",
                                "uniqueItems": True,
                            },
                        },
                    },
                    "additionalProperties": False,
                },
            },
            "additionalProperties": False,
        },
    },
    "additionalProperties": False,
}


def to_list(s, sort=True):
    if not s:
        return []
    if sort:
        return list(sorted(filter(None, (l.strip() for l in s.splitlines()))))
    else:
        return list(filter(None, (l.strip() for l in s.splitlines())))


class ConfError(ValueError):
    pass


class IniConf:
    def __init__(
        self,
        *,
        summary: Optional[str] = None,
        supported_arches: Optional[list[str]] = None,
        all_features: Optional[bool] = None,
        unwanted_features: Optional[list[str]] = None,
        enabled_features: Optional[list[str]] = None,
        buildrequires: Optional[list[str]] = None,
        testrequires: Optional[list[str]] = None,
        bin_requires: Optional[list[str]] = None,
        lib_requires: Optional[dict[Optional[str], list[str]]] = None,
        conf_file: bool = False,
    ):
        self.summary: Optional[str] = summary
        self.supported_arches: list[str] = supported_arches or list()
        self.all_features: Optional[bool] = all_features
        self.unwanted_features: list[str] = unwanted_features or list()
        self.enabled_features: list[str] = enabled_features or list()
        self.buildrequires: list[str] = buildrequires or list()
        self.testrequires: list[str] = testrequires or list()
        self.bin_requires: list[str] = bin_requires or list()
        self.lib_requires: dict[Optional[str], list[str]] = lib_requires or dict()
        self.conf_file: bool = conf_file

    def __eq__(self, other):
        if not isinstance(other, IniConf):
            return False  # pragma nocover

        # order of items in the list is consistent:
        # lists are sorted when parsing rust2rpm.conf files
        return (
            self.summary == other.summary
            and self.supported_arches == other.supported_arches
            and self.all_features == other.all_features
            and self.unwanted_features == other.unwanted_features
            and self.enabled_features == other.enabled_features
            and self.buildrequires == other.buildrequires
            and self.testrequires == other.testrequires
            and self.bin_requires == other.bin_requires
            and self.lib_requires == other.lib_requires
        )

    @staticmethod
    def load(filenames: list[str], target: str, features: set[str]) -> "IniConf":
        conf = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())
        confs = conf.read(filenames)

        if len(confs) > 1:  # pragma nocover
            raise FileExistsError

        # clean up configuration files with deprecated names
        if ".rust2rpm.conf" in confs:  # pragma nocover
            os.rename(".rust2rpm.conf", "rust2rpm.conf")
            log.info("Renamed deprecated, hidden .rust2rpm.conf file to rust2rpm.conf.")

        if "_rust2rpm.conf" in confs:  # pragma nocover
            os.rename("_rust2rpm.conf", "rust2rpm.conf")
            log.info("Renamed deprecated _rust2rpm.conf file to rust2rpm.conf.")

        # merge target-specific configuration with default configuration
        if target not in conf:
            conf.add_section(target)
        merged = conf[target]

        # validate configuration file
        valid_targets = ["fedora", "mageia", "openmandriva", "opensuse", "plain"]
        valid_keys = [
            "summary",
            "supported-arches",
            "all-features",
            "unwanted-features",
            "enabled-features",
            "buildrequires",
            "testrequires",
            "lib.requires",
            "bin.requires",
        ]

        # the "default" feature is always implicitly defined
        if "default" not in features:
            features.add("default")

        for feature in features:
            valid_keys.append(f"lib+{feature}.requires")

        # check section names
        for section in conf.keys():
            if section != "DEFAULT" and section not in valid_targets:
                raise ConfError(f"Invalid section: {section!r}")

        # check setting keys
        for key in merged.keys():
            if key not in valid_keys:
                raise ConfError(f"Invalid key: {key!r}")

        # parse configuration and validate settings
        settings = dict()

        if summary := merged.get("summary"):
            settings["summary"] = summary

        if supported_arches := merged.get("supported-arches"):
            settings["supported_arches"] = to_list(supported_arches, False)

        if merged.get("all-features"):
            all_features = merged.getboolean("all-features")
            settings["all_features"] = all_features

        if unwanted_features := merged.get("unwanted-features"):
            settings["unwanted_features"] = to_list(unwanted_features)

            for unwanted_feature in settings["unwanted_features"]:
                if unwanted_feature not in features:
                    raise ConfError(f'Unrecognized "unwanted" feature: {unwanted_feature!r}')

        if enabled_features := merged.get("enabled-features"):
            settings["enabled_features"] = to_list(enabled_features)

            for enabled_feature in settings["enabled_features"]:
                if enabled_feature not in features:
                    raise ConfError(f'Unrecognized "enabled" feature: {enabled_feature!r}')

        if buildrequires := merged.get("buildrequires"):
            settings["buildrequires"] = to_list(buildrequires)
        if testrequires := merged.get("testrequires"):
            settings["testrequires"] = to_list(testrequires)
        if bin_requires := merged.get("bin.requires"):
            settings["bin_requires"] = to_list(bin_requires)

        if lib_requires := merged.get("lib.requires"):
            if "lib_requires" not in settings.keys():
                settings["lib_requires"] = dict()
            settings["lib_requires"][None] = to_list(lib_requires)  # type: ignore

        for feature in features:
            if lib_feature_requires := merged.get(f"lib+{feature}.requires"):
                if "lib_requires" not in settings.keys():
                    settings["lib_requires"] = dict()
                settings["lib_requires"][feature] = to_list(lib_feature_requires)  # type: ignore

        settings["conf_file"] = len(confs) == 1

        return IniConf(**settings)  # type: ignore


class TomlConf:
    def __init__(self, data: Optional[dict] = None):
        self._data = data or dict()

    def __repr__(self) -> str:
        return repr(self._data)

    def __eq__(self, other) -> bool:
        if not isinstance(other, TomlConf):
            return False  # pragma nocover

        return self._data == other._data

    @property
    def _package(self) -> Optional[dict]:
        return self._data.get("package")

    @property
    def package_summary(self) -> Optional[str]:
        if package := self._package:
            return package.get("summary")
        else:
            return None

    @property
    def package_description(self) -> Optional[str]:
        if package := self._package:
            return package.get("description")
        else:
            return None

    @property
    def package_supported_arches(self) -> Optional[list[str]]:
        if package := self._package:
            return package.get("supported-arches")
        else:
            return None

    @property
    def _tests(self) -> Optional[dict[str, Any]]:
        return self._data.get("tests")

    @property
    def tests_run(self) -> Optional[list[str]]:
        if tests := self._tests:
            return tests.get("run")
        return None

    @property
    def tests_skip(self) -> Optional[list[str]]:
        if tests := self._tests:
            return tests.get("skip")
        return None

    @property
    def tests_skip_exact(self) -> Optional[bool]:
        if tests := self._tests:
            return tests.get("skip-exact")
        return None

    @property
    def tests_comments(self) -> Optional[list[str]]:
        if tests := self._tests:
            return tests.get("comments")
        return None

    @property
    def _features(self) -> Optional[dict]:
        return self._data.get("features")

    @property
    def features_enable_all(self) -> Optional[bool]:
        if features := self._features:
            return features.get("enable-all")
        else:
            return None

    @property
    def features_enable(self) -> Optional[list[str]]:
        if features := self._features:
            return features.get("enable")
        else:
            return None

    @property
    def features_hide(self) -> Optional[list[str]]:
        if features := self._features:
            return features.get("hide")
        else:
            return None

    @property
    def _requires(self) -> Optional[dict]:
        return self._data.get("requires")

    @property
    def requires_build(self) -> Optional[list[str]]:
        if requires := self._requires:
            return requires.get("build")
        else:
            return None

    @property
    def requires_test(self) -> Optional[list[str]]:
        if requires := self._requires:
            return requires.get("test")
        else:
            return None

    @property
    def requires_lib(self) -> Optional[list[str]]:
        if requires := self._requires:
            return requires.get("lib")
        else:
            return None

    @property
    def requires_bin(self) -> Optional[list[str]]:
        if requires := self._requires:
            return requires.get("bin")
        else:
            return None

    @property
    def requires_features(self) -> Optional[dict[str, list[str]]]:
        if requires := self._requires:
            return requires.get("features")
        else:
            return None

    @staticmethod
    def load(path: str, features: set[str]) -> "TomlConf":
        with open(path, "rb") as file:
            toml = tomllib.load(file)

        jsonschema.validate(toml, TOML_SCHEMA)

        if tests := toml.get("tests"):
            if run := tests.get("run"):
                if "all" in run:
                    if len(run) != 1:
                        raise ConfError(f"Invalid set of tests to run: {run!r}")
                if "none" in run:
                    if len(run) != 1:
                        raise ConfError(f"Invalid set of tests to run: {run!r}")

        # the "default" feature is always implicitly defined
        if "default" not in features:
            features.add("default")

        if features_list := toml.get("features"):
            if enable_list := features_list.get("enable"):
                for enabled in enable_list:
                    if enabled not in features:
                        raise ConfError(f'Unrecognized "enabled" feature: {enabled!r}')
            if hide_list := features_list.get("hide"):
                for hidden in hide_list:
                    if hidden not in features:
                        raise ConfError(f'Unrecognized "hidden" feature: {hidden!r}')

        if requires_table := toml.get("requires"):
            if features_list := requires_table.get("features"):
                for enabled in features_list:
                    if enabled not in features:
                        raise ConfError(f'Unrecognized "Requires" for feature: {enabled!r}')

        return TomlConf(toml)

    @staticmethod
    def from_conf(conf: IniConf) -> "TomlConf":
        settings: dict[str, Any] = dict()

        package: dict[str, Any] = dict()
        if conf.summary:
            package["summary"] = conf.summary
        if conf.supported_arches:
            package["supported-arches"] = conf.supported_arches
        if package:
            settings["package"] = package

        features: dict[str, Any] = dict()
        if conf.all_features is not None:
            features["enable-all"] = conf.all_features
        if conf.enabled_features:
            features["enable"] = conf.enabled_features
        if conf.unwanted_features:
            features["hide"] = conf.unwanted_features
        if features:
            settings["features"] = features

        requires: dict[str, Any] = dict()
        if conf.buildrequires:
            requires["build"] = conf.buildrequires
        if conf.testrequires:
            requires["test"] = conf.testrequires
        if conf.bin_requires:
            requires["bin"] = conf.bin_requires
        if conf.lib_requires:
            for key, value in conf.lib_requires.items():
                if key is None:
                    requires["lib"] = value
                else:
                    if "features" not in requires.keys():
                        requires["features"] = dict()
                    requires["features"][key] = value
        if requires:
            settings["requires"] = requires

        jsonschema.validate(settings, TOML_SCHEMA)

        return TomlConf(settings)


def load_config(features: set[str], target: str) -> TomlConf:
    # attempt to load rust2rpm.toml
    try:
        tomlconf = TomlConf.load("rust2rpm.toml", features)

    except FileNotFoundError as exc:
        tomlconf = None

    except tomllib.TOMLDecodeError as exc:
        log.error("Cannot read rust2rpm.toml file (TOML syntax error):")
        log.error(str(exc))
        sys.exit(1)

    except jsonschema.ValidationError as exc:
        if not exc.path:
            log.error("Invalid rust2rpm.toml file (unknown setting or table):")
            log.error(exc.message)
        else:
            err_path = ""
            for elem in exc.path:
                if isinstance(elem, int):
                    err_path += f"[{elem}]"
                else:
                    err_path += f"/{elem}"
            log.error(f"Invalid rust2rpm.toml file (invalid setting at {err_path!r}):")
            log.error(exc.message)
        sys.exit(1)

    except ConfError as exc:
        log.error("Invalid rust2rpm.toml file:")
        log.error(str(exc))
        sys.exit(1)

    if tomlconf:
        return tomlconf

    # fall back to rust2rpm.conf and translate
    try:
        # known file names (only in the current working directory)
        filenames = [".rust2rpm.conf", "_rust2rpm.conf", "rust2rpm.conf"]
        distconf = IniConf.load(filenames, target, features)

    except FileExistsError:
        log.error(
            "More than one *rust2rpm.conf file is present in this directory. "
            + "Ensure that there is only one, and that it has the correct contents."
        )
        sys.exit(1)

    except ConfError as exc:
        log.error("Invalid rust2rpm configuration file:")
        log.error(str(exc))
        sys.exit(1)

    if distconf.conf_file:
        log.warn(
            "The rust2rpm.conf file format is deprecated and support for it will "
            + "be removed in a future release of rust2rpm. Please migrate settings "
            + "to the new TOML based format (rust2rpm.toml)."
        )

    tomlconf = TomlConf.from_conf(distconf)
    return tomlconf


def conf_to_bcond_check(conf: TomlConf) -> bool:
    if tests_run := conf.tests_run:
        return "none" not in tests_run
    else:
        return True


def conf_to_test_comments(conf: TomlConf) -> list[str]:
    if conf.tests_comments:
        comments = conf.tests_comments
    else:
        return []

    lines = []
    for comment in comments:
        lines.extend(
            textwrap.wrap(
                comment,
                width=80,
                initial_indent="# * ",
                subsequent_indent="#   ",
                break_long_words=False,
                break_on_hyphens=False,
            )
        )

    return lines


def conf_to_cargo_test_args(conf: TomlConf) -> list[str]:
    if skip := conf.tests_skip:
        skip_flags = []
        for name in skip:
            skip_flags.extend(["--skip", name])
    else:
        skip_flags = []

    if conf.tests_skip_exact is True:
        skip_exact_flag = ["--exact"]
    else:
        skip_exact_flag = []

    if run := conf.tests_run:
        if not ("all" in run or "none" in run):
            multi = []
            for kind in run:
                if skip_flags:
                    multi.append(" " + " ".join(["--", f"--{kind}", "--"] + skip_exact_flag + skip_flags))
                else:
                    multi.append(" " + " ".join(["--", f"--{kind}"]))
            return multi

    if skip_flags:
        return [" " + " ".join(["--", "--"] + skip_exact_flag + skip_flags)]
    else:
        return []
